//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

// RTC Near Field Communications
#include "RTCMegaMaster.h"
#include "RTC_NFC.h"
#include "Engine_Commands.h"
#include "Train_Control.h"
#include "Comm_Thread.h"
#include "Debug_Controls.h"
#include "Clipbrd.hpp"
#include "EngineListPopOut.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma link "HideButton"
#pragma resource "*.dfm"
TPopOutEngineList *PopOutEngineList;

extern TOp * ptrOp[ABSOLUTEMAX];   		// Operation Control Pointers indexed by actual engine number because 0 is a factory reset engine

extern TCommThread * ptrTIUCommThread;	// pointer to TIU communications thread

//---------------------------------------------------------------------------
__fastcall TPopOutEngineList::TPopOutEngineList(TComponent* Owner)
    : TForm(Owner)
{
RegFileRootKey =  MainForm->RegFileRootKey;
//
TRegIniFile * const RegFile = new TRegIniFile(RegFileRootKey);
assert(RegFile != NULL);
__try {
    // read window position
    Top = RegFile->ReadInteger("EngineList", "Top", Top);
    Height = RegFile->ReadInteger("EngineList", "Height", Height);
    Left = RegFile->ReadInteger("EngineList", "Left", Left);

//    EngineData->Selection.Top = RegFile->ReadInteger("EngineList", "SelTop", EngineData->Selection.Top);
//    EngineData->Selection.Bottom = RegFile->ReadInteger("EngineList", "SelBottom", EngineData->Selection.Bottom);
    // if using multiple monitors and one was removed, pull the window
    // back onto the remaining monitor
    int MC = Screen->MonitorCount;
    if (MC == 1) {
        int Wth = Screen->Monitors[0]->Width;
		if (Left > Wth) Left -= Wth;
        int Hth = Screen->Monitors[0]->Height;
        if (Top > Hth) Top -= Hth;
        }
	}
__finally {
	// Deallocate RegFile
	delete RegFile;
	}
//
EngineData->DefaultColWidth = 30;		// default
EngineData->ColWidths[ENG_NAME] = EngineData->Width - EngineData->DefaultColWidth;	// lots of space for the engine name
}
//---------------------------------------------------------------------------

__fastcall TPopOutEngineList::~TPopOutEngineList(void)
{
TRegIniFile * const RegFile = new TRegIniFile(RegFileRootKey);
assert(RegFile != NULL);
__try {	    // erase EngineList key
    RegFile->EraseSection("EngineList");
    // save window position
    RegFile->WriteInteger("EngineList", "Top", Top > 0 ? Top : 0);
    RegFile->WriteInteger("EngineList", "Height", Height);
    RegFile->WriteInteger("EngineList", "Left", Left);
    }
__finally {
    // Deallocate RegFile
    delete RegFile;
    }
}
//---------------------------------------------------------------------------

void __fastcall TPopOutEngineList::PopoutEngineDataDrawCell(TObject *Sender,
      int ACol, int ARow, TRect &Rect, TGridDrawState State)
{
// Draw the contents of each cell
EngineData->Canvas->Font->Size = 12;
// http://bcbjournal.org/articles/vol3/9904/Dealing_with_sets.htm?PHPSESSID=2f96cbddf92baa7dc77f7a33591f59b5
EngineData->Canvas->Font->Style = EngineData->Canvas->Font->Style >> fsItalic;
// if the cell is selected, highlight it
if( State.Contains(gdSelected) ) {
    EngineData->Canvas->Font->Color = clWhite;
	EngineData->Canvas->Brush->Color = clBlue;		// White on Blue
	} else {
    if (ptrOp[ARow] != NULL) {
		EngineData->Canvas->Font->Color = clBlack;	// Started up Engines are Black on Lime
		EngineData->Canvas->Brush->Color = clLime;
    	} else {
		EngineData->Canvas->Font->Color = clBlack;	// Normal Black on White
		EngineData->Canvas->Brush->Color = clWhite;
        }
	}
if ((ARow != 0) && (ARow != ALLENGINE) && (ACol == ENG_NAME) && (EngineData->Cells[ENG_STATE][ARow] == INACTIVE)) {
	// if Engine is inactive make the cell Italic.
    // Skip the column title (Row 0) and All Engine Operation (Row 100)
    EngineData->Canvas->Font->Style = EngineData->Canvas->Font->Style << fsItalic;
	}
// move a little to the right
Rect.Left = Rect.Left + LISTLEFTOFFSET;
// fill the rectangle
EngineData->Canvas->FillRect(Rect);
// draw the text
String text;
//if (DEBUGGING && (ACol == ENG_NAME) && (ARow != ENG_LABEL_ROW) && (ARow != ENG_ALLROW) && (EngineData->Cells[ENG_NAME][ARow] != "*") &&
//        (EngineData->Cells[ENG_STATE][ARow] == ACTIVE)) {
if ((ACol == ENG_NAME) && (ARow != ENG_LABEL_ROW) && (ARow != ENG_ALLROW) && (EngineData->Cells[ENG_NAME][ARow] != "*")) {
    text = EngineData->Cells[ACol][ARow] + "  TIU" + String(EngineData->Cells[TIU_NUMBER][ARow]);
    } else {
    text = EngineData->Cells[ACol][ARow];
    }
EngineData->Canvas->TextRect(Rect, Rect.Left, Rect.Top, text);
}
//---------------------------------------------------------------------------

void __fastcall TPopOutEngineList::PopoutEngineDataClick(TObject *Sender)
{
// Reflect new row to MainForm->EngineData
if (MainForm && (MainForm->EngineData->Row != EngineData->Row)) MainForm->EngineData->Row = EngineData->Row;
}
//---------------------------------------------------------------------------

void __fastcall TPopOutEngineList::SetConsistState(int Consist, int State)
{
// set the active/inactive state of an Engine
//
assert((State == ACTIVE) || (State == INACTIVE));
if (Consist == 0) return;	// factory reset engine not allowed
assert((Consist >= 1) && (Consist <= ENGINEMAX));	// 1-99
if (EngineData->Cells[ENG_STATE][Consist] == ACTIVE) {		// Adjust Counts
	if (State == INACTIVE) {
        MainForm->ActiveEngineCount->Value--;
		MainForm->InactiveEngineCount->Value++;
    	}
	}
if (EngineData->Cells[ENG_STATE][Consist] == INACTIVE) {	// Adjust Counts
	if (State == ACTIVE) {
        MainForm->ActiveEngineCount->Value++;
		MainForm->InactiveEngineCount->Value--;
    	}
	}

EngineData->Cells[ENG_STATE][Consist] = State;	// set active or inactive
MainForm->EngineData->Cells[ENG_STATE][Consist] = State;	// set active or inactive
// redraw the control
EngineData->Repaint();
MainForm->EngineData->Repaint();
return;
}
//---------------------------------------------------------------------------

void __fastcall TPopOutEngineList::SetEnginesActiveClick(TObject *Sender)
{
for (int EngineNo = EngineData->Selection.Top; EngineNo <= EngineData->Selection.Bottom ; EngineNo++) {
    if (EngineData->Cells[ENG_NAME][EngineNo] != "*") {
        SetConsistState(EngineNo, ACTIVE);
        }
    }
// update users of the Engine List
MainForm->UpdateEngDataUsers();
}
//---------------------------------------------------------------------------

void __fastcall TPopOutEngineList::SetEnginesInactiveClick(TObject *Sender)
{
for (int EngineNo = EngineData->Selection.Top; EngineNo <= EngineData->Selection.Bottom ; EngineNo++) {
    if (EngineData->Cells[ENG_NAME][EngineNo] != "*") {
        SetConsistState(EngineNo, INACTIVE);		// set inactive
        }
    }
// update users of the Engine List
MainForm->UpdateEngDataUsers();
}
//---------------------------------------------------------------------------

void __fastcall TPopOutEngineList::SetEnginesintoCommandModeClick(
      TObject *Sender)
{
RemoteNo = MainForm->RemoteSpin->ValueAsInt;
// if TIU Port is not connected fail
if (!MainForm->IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }
for (int EngineNo = EngineData->Selection.Top; EngineNo <= EngineData->Selection.Bottom ; EngineNo++) {
    if (EngineData->Cells[ENG_NAME][EngineNo] != "*" && EngineData->Cells[ENG_STATE][EngineNo] == ACTIVE) {
        (void)ptrTIUCommThread->Send_String_FIFO(EngineNo, "m4", EngineData->Cells[TIU_NUMBER][EngineNo].ToInt(), RemoteNo, 11, SHORT_WAIT, RECORD_OK);
        }
    }
}
//---------------------------------------------------------------------------

void __fastcall TPopOutEngineList::SetEnginesintoConventionalModeClick(
      TObject *Sender)
{
RemoteNo = MainForm->RemoteSpin->ValueAsInt;
// if TIU Port is not connected fail
if (!MainForm->IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }
for (int EngineNo = EngineData->Selection.Top; EngineNo <= EngineData->Selection.Bottom ; EngineNo++) {
    if (EngineData->Cells[ENG_NAME][EngineNo] != "*" && EngineData->Cells[ENG_STATE][EngineNo] == ACTIVE) {
        (void)ptrTIUCommThread->Send_String_FIFO(EngineNo, "m2", EngineData->Cells[TIU_NUMBER][EngineNo].ToInt(), RemoteNo, 11, SHORT_WAIT, RECORD_OK);
        }
    }
}
//---------------------------------------------------------------------------

void __fastcall TPopOutEngineList::SetEnginesintoLegacyModeClick(
      TObject *Sender)
{
RemoteNo = MainForm->RemoteSpin->ValueAsInt;
// if TIU Port is not connected fail
if (!MainForm->IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }
for (int EngineNo = EngineData->Selection.Top; EngineNo <= EngineData->Selection.Bottom ; EngineNo++) {
    if (EngineData->Cells[ENG_NAME][EngineNo] != "*" && EngineData->Cells[ENG_STATE][EngineNo] == ACTIVE) {
        (void)ptrTIUCommThread->Send_String_FIFO(EngineNo, "m8", EngineData->Cells[TIU_NUMBER][EngineNo].ToInt(), RemoteNo, 11, SHORT_WAIT, RECORD_OK);
        }
    }
}
//---------------------------------------------------------------------------

void __fastcall TPopOutEngineList::SetEnginesPS2Click(TObject *Sender)
{
for (int EngineNo = EngineData->Selection.Top; EngineNo <= EngineData->Selection.Bottom ; EngineNo++) {
    if (EngineData->Cells[ENG_NAME][EngineNo] != "*") {
        SetEnginesPS2->Checked = true;
        SetEnginesPS3->Checked = false;
        EngineData->Cells[ENG_PS][EngineNo] = PS2;
        MainForm->SetEnginePS2->Checked = true;
        MainForm->SetEnginePS3->Checked = false;
        MainForm->EngineData->Cells[ENG_PS][EngineNo] = PS2;
        }
    }
}
//---------------------------------------------------------------------------

void __fastcall TPopOutEngineList::SetEnginesPS3Click(TObject *Sender)
{
for (int EngineNo = EngineData->Selection.Top; EngineNo <= EngineData->Selection.Bottom ; EngineNo++) {
    if (EngineData->Cells[ENG_NAME][EngineNo] != "*") {
        SetEnginesPS2->Checked = false;
        SetEnginesPS3->Checked = true;
        EngineData->Cells[ENG_PS][EngineNo] = PS3;
        MainForm->SetEnginePS2->Checked = false;
        MainForm->SetEnginePS3->Checked = true;
        MainForm->EngineData->Cells[ENG_PS][EngineNo] = PS3;
        }
    }
}
//---------------------------------------------------------------------------

void __fastcall TPopOutEngineList::StartUpSelectedEngines1Click(
      TObject *Sender)
{
// if TIU Port is not connected fail
if (!MainForm->IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }
for (int EngineNo = EngineData->Selection.Top; EngineNo <= EngineData->Selection.Bottom ; EngineNo++) {
    if (EngineData->Cells[ENG_NAME][EngineNo] != "*") {
        MainForm->EngineData->Row = EngineNo;
        MainForm->StartupButtonClick(Sender);
        }
    }
}
//---------------------------------------------------------------------------

void __fastcall TPopOutEngineList::ShutdownSelectedEngines1Click(
      TObject *Sender)
{
// if TIU Port is not connected fail
if (!MainForm->IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }
for (int EngineNo = EngineData->Selection.Top; EngineNo <= EngineData->Selection.Bottom ; EngineNo++) {
    if (EngineData->Cells[ENG_NAME][EngineNo] != "*") {
        // You can shutdown an engine even it was not started up
        MainForm->EngineData->Row = EngineNo;
        MainForm->ShutdownButtonClick(Sender);        
        }
    }
}
//---------------------------------------------------------------------------

void __fastcall TPopOutEngineList::HideButton1Click(TObject *Sender)
{
Visible = false;    
}
//---------------------------------------------------------------------------

void __fastcall TPopOutEngineList::PopoutEngineDataDblClick(TObject *Sender)
{
if (EngineData->Cells[ENG_NAME][EngineNo] != "*") {
    MainForm->EngineDataDblClick(Sender);
    }
}
//---------------------------------------------------------------------------

void __fastcall TPopOutEngineList::EngineDataKeyDown(TObject *Sender,
      WORD &Key, TShiftState Shift)
{
// if ENTER is pressed, startup the selected engines
if (Key == VK_RETURN) {
	if (EngineData->Cells[ENG_NAME][EngineData->Row] != "*") {
		StartUpSelectedEngines1Click(Sender);
   		}
    }
}
//---------------------------------------------------------------------------

void __fastcall TPopOutEngineList::CopyrightClick(TObject *Sender)
{
// Copy the Form to the clipboard.
// In BCB4.0, there is no SetSize method associated with TBitmap
// so we have to use Width and Height separately.
//
// see:
// https://stackoverflow.com/questions/3190777/how-do-i-copy-a-form-as-an-image-to-the-clipboard
//
Graphics::TBitmap * bm;
bm = new Graphics::TBitmap;
try
	{
    #if __BORLANDC__ >= 0x720
	// Defined for CB10.1 C++ programs that use the VCL.
	bm->SetSize(ClientWidth, ClientHeight);
    #else
    bm->Width = ClientWidth;
    bm->Height = ClientHeight;
    #endif
	BitBlt(bm->Canvas->Handle, 0, 0, ClientWidth, ClientHeight, Canvas->Handle, 0, 0, SRCCOPY);
    Clipboard()->Assign(bm);
    }
__finally
	{
	delete bm;
    }    
}
//---------------------------------------------------------------------------

void __fastcall TPopOutEngineList::AllEngineButtonClick(TObject *Sender)
{
//
// Select Row 100 for All Engine
EngineData->Row = ALLENGINE;
if (MainForm->EngineData->Row != EngineData->Row) MainForm->EngineData->Row = EngineData->Row;
}
//---------------------------------------------------------------------------

void __fastcall TPopOutEngineList::RefreshClick(TObject *Sender)
{
//
// cannot assign a TStringGrid to a TStringGrid : PopOutEngineList->EngineData->Assign(EngineData);
for (int I = 0; I < EngineData->ColCount; I++)
    EngineData->Cols[I]->Assign(MainForm->EngineData->Cols[I]);
EngineData->Row = MainForm->EngineData->Row;
EngineData->Col = MainForm->EngineData->Col;
Engine_List_Label->Caption = MainForm->Engine_List_Label->Caption;
}
//---------------------------------------------------------------------------
